import java.awt.*;
import java.io.*;

/**
 * An abstract class that stores the representation of an active piece 
 * including position, rotation, color as well as the model of how the
 * shape should look in all of its rotations
 */
public abstract class Piece implements Serializable
{
	// Instance variables
	protected PlayingField field;
	protected int xPosition, yPosition;
	protected int rotation;
	protected Color color;
	protected int[][][] model;

	/**
	 * Constructor. Creates a new instance of piece given a position, color and model.
	 * @param xPosition The x position of the piece
	 * @param yPosition The y position of the piece
	 * @param color The color of the piece
	 * @param model The rotation and shape mode for the piece
	 */
	public Piece(int xPosition, int yPosition, Color color, int[][][] model, PlayingField field)
	{
		// Sets component fields
		this.xPosition = xPosition;
		this.yPosition = yPosition;
		this.color = color;
		this.model = model;
		this.field = field;
		rotation = 0;
	}
	
	/**
	 * Getter method for the location of the Piece
	 * @return A Position representing the position of the piece
	 */
	public Point getLocation()
	{
		return (new Point(xPosition, yPosition));
	}

	/**
	 * This method will attempt to move a piece to a given location on the field. If the 
	 * move produces a collision of some kind, it will not be carried out. If the move
	 * is sucessful, it returns true, false if the move failed.
 	 * @param location The location to attempt to move the piece.
 	 * @return true if the move was sucessful
 	 */
	public boolean move(Point location)
	{
		// Get the current configuration
		int[][] curPosition = model[rotation];
		
		// Go through each part of the Piece
		for (int yOffset = 0; yOffset < curPosition.length; yOffset++)
		{
			for (int xOffset = 0; xOffset < curPosition[0].length; xOffset++)
			{
				if (curPosition[yOffset][xOffset] == 1)
				{
					// Get actual x and y values for that position 
					int newX = location.x + xOffset;
					int newY = location.y - yOffset;
					
					// Return with collision if the Piece is out of range
					if (!field.isTileInRange(newX, newY))
						return false;
					
					// If it overlaps with another tile return with collision
					if (!field.isTileVacant(newX, newY))
						return false;
				}
			}
		}
		
		// If it hasn't returned false yet, move must be legal
		xPosition = location.x;
		yPosition = location.y;
		return true;
	}

	/**
	 * This method will attempt to rotate the piece.
	 * @return true if the rotation was successful, false if the rotation would have resulted in collision
	 */
	public boolean rotate()
	{
		// Get the next configuration
		int newRotation = (rotation + 1) % model.length;
		int[][] newPosition = model[newRotation];
		
		// Go through each part of the Piece
		for (int yOffset = 0; yOffset < newPosition.length; yOffset++)
		{
			for (int xOffset = 0; xOffset < newPosition[0].length; xOffset++)
			{
				if (newPosition[yOffset][xOffset] == 1)
				{
					// Get actual x and y values for that position 
					int newX = xPosition + xOffset;
					int newY = yPosition - yOffset;
					
					// Return with collision if the Piece is out of range
					if (!field.isTileInRange(newX, newY))
						return false;
					
					// If it overlaps with another tile return with collision
					if (!field.isTileVacant(newX, newY))
						return false;
				}
			}
		}
		
		// If it hasn't returned false yet, rotation must be legal
		rotation = newRotation;
		return true;

	}

	/**
	 * This method will fuse the piece to the board, by filling in the board tiles
	 * that the piece overlaps with. 
	 */
	public void settle()
	{
		// Get the current configuration
		int[][] curPosition = model[rotation];
		
		// Go through each part of the Piece
		for (int yOffset = 0; yOffset < curPosition.length; yOffset++)
		{
			for (int xOffset = 0; xOffset < curPosition[0].length; xOffset++)
			{
				if (curPosition[yOffset][xOffset] == 1)
				{
					// Get actual x and y values for that position 
					int newX = xPosition + xOffset;
					int newY = yPosition - yOffset;
					
					// Fill in the tile
					field.fillTile(newX, newY, color);
				}
			}
		}
	}

	/**
	 * This method draws the piece on the game board. 
	 * @param g The Graphics object
	 * @param tileDimension The dimension of the tile.
	 */
	public void draw(Graphics g, Dimension tileDimension)
	{
		// Get the current configuration
		int[][] curPosition = model[rotation];
		
		// Go through each part of the Piece
		for (int yOffset = 0; yOffset < curPosition.length; yOffset++)
		{
			for (int xOffset = 0; xOffset < curPosition[0].length; xOffset++)
			{
				if (curPosition[yOffset][xOffset] == 1)
				{
					// Get actual x and y values for that position 
					int newX = xPosition + xOffset;
					int newY = yPosition - yOffset;
					
					// Draw the tile
					field.drawTile(g, tileDimension, newX, newY, color);
				}
			}
		}
	}
	
	/**
	 * This method draws the piece on a panel at an arbitrary location  
	 * @param g The Graphics object
	 * @param tileDimension The dimension of the tile.
	 */
	public void drawSimple(Graphics g, Dimension tileDimension, int x, int y)
	{
		// Get the current configuration
		int[][] curPosition = model[rotation];
		
		// Go through each part of the Piece
		for (int yOffset = 0; yOffset < curPosition.length; yOffset++)
		{
			for (int xOffset = 0; xOffset < curPosition[0].length; xOffset++)
			{
				if (curPosition[yOffset][xOffset] == 1)
				{
					// Draw the current tile
					g.setColor(color);
					g.fillRect(x+xOffset*tileDimension.width, y+yOffset*tileDimension.height, tileDimension.width, tileDimension.height);
				}
			}
		}
	}
	
  	/**
	 * This method allows for the object to be read with serialization
	 * @param aInputStream The input stream to use
	 */
	private void readObject(ObjectInputStream aInputStream) throws IOException, ClassNotFoundException
	{
     aInputStream.defaultReadObject();
  	}
	
	/**
	 * This method allows for the object to be written with serialization
	 * @param aOutputStream The output stream to use
	 */
	private void writeObject(ObjectOutputStream aOutputStream) throws IOException 
	{
      aOutputStream.defaultWriteObject();
   }
}