import java.awt.*;
import java.io.*;

/**
 * This class stores the state of a Tertis board, checks for lines to
 * clear, and checks for a loss condition.
 */
public class PlayingField implements Serializable
{
	// Instance varaibles
	private int rows;
	private int cols;
	private int freeZone;
	private Color[][] board;
	
	/**
	 * Constructor. Creates new instance of PlayingField.
	 * @param rows The y dimension of the PlayingField
	 * @param cols The x dimension of the PlayingField
	 * @param freeZone The y distance above the field where the Pieces spawn  
	 */
	public PlayingField(int rows, int cols, int freeZone)
	{
		// Instantiate instance varaibles
		this.rows = rows;
		this.cols = cols;
		this.freeZone = freeZone;
		
		// Create a new empty board
		board = new Color[rows][cols];	
	}
	 
	/**
	 * Scans all of the rows, such that when a row is completely fileed with colors,
	 * the tow is cleared. All the tiles above that row are moved down one row. Keeps
	 * track of the number of rows cleared.
	 * @return Number of rows cleared
	 */
	public int checkForClears()
	{
		// Start clear counter
		int rowsCleared = 0;
		// Check each row
		for (int row = 0; row < rows - freeZone; row++)
		{
			// Check each column on the row
			boolean rowFilled = true;
			for (int col = 0; col < cols; col++)
			{
				// Make rowFilled false if there is a null value
				if (board[row][col] == null)
				{
					rowFilled = false;
					break;
				}
			}
			
			// If the row was full
			if (rowFilled)
			{
				// Increment counter
				rowsCleared++;
				// Shift each row down
				for (int newRow = row; newRow < rows - freeZone; newRow++)
				{
					for (int col = 0; col < cols; col++)
					{
						board[newRow][col] = board[newRow+1][col];
					}
				}
				// Stay at the same row next cycle
				row--;		
			}
		}
		return rowsCleared;
	}
	
	/**
	 * If the top row of the playing field has any colors, the player loses the game.
	 * Returns true if the player has lost.
	 * @return true if the player has lost
	 */
	public boolean isLossConditionMet()
	{
		// First row in free zone
		int row = rows - freeZone;
		
		// Looks at each column to see if there is a piece there
		boolean hasLost = false;
		for (int col = 0; col < cols; col++)
		{
			if (board[row][col] != null)
			{
				hasLost = true;
				break;
			}
		}
		return hasLost;
	}
	
	/**
	 * Checks to see if a specific tile at x and y is vacant.
	 * @param x The x location of the tile
	 * @param y The y location of the tile
	 * @return true if the tile is vacant
	 */
	public boolean isTileVacant(int x, int y)
	{
		// If it's null, it's vacant, else it's not vacant
		boolean isVacant;
		if (board[y][x] == null)
			isVacant = true;
		else
			isVacant = false;
		return isVacant;
	}
	
	/** 
	 * Checks to make sure that the given x and y position is in the
	 * parameters of the board.
	 * @param x The x location of the tile
	 * @param y The y location of the tile
	 * @return true if the location is in range
	 */
	public boolean isTileInRange(int x, int y)
	{
		// Return false if it's less than 0, or greater than the
		// number or columns or rows for x and y resp
		if (x < 0 || x >= cols || y < 0 || y >= rows)
			return false;
		else
			return true;
	}
	
	/**
	 * Sets the contents of a particular tile.
	 * @param x The x location of the tile
	 * @param y The y location of the tile
	 * @param color The color of the tile
	 */
	public void fillTile(int x, int y, Color color)
	{
		board[y][x] = color;
	}
	
	/**
	 * Renders all of the tiles properly on the screen.
	 * @param g The Graphics object
	 * @param tileDimension The dimension of the tile
	 */
	public void draw(Graphics g, Dimension tileDimension)
	{	
		// Go through each tile and seperately paint it
		for (int row = 0; row < rows; row++)
		{
			for (int col = 0; col < cols; col++)
			{
				drawTile(g, tileDimension, col, row, board[row][col]);
			}
		}
	}
	
	/**
	 * Draws a single tile on the screen of a given color.
	 * @param g The Graphics object
	 * @param tileDimension The dimension of the tile
	 */
	public void drawTile(Graphics g, Dimension tileDimension, int x, int y, Color color)
	{
		// Do nothing if it's an empty space
		if (color == null)
			return;
		
		// Find x and y offsets
		int xOffset = x*tileDimension.width;
		int yOffset = ((rows-1)* tileDimension.height) - y*tileDimension.height;
		
		// Paint starting at the offset for the dimension of the tile
		g.setColor(color);
		g.fillRect(xOffset, yOffset, tileDimension.width, tileDimension.height);
	}
	
	/**
	 * This method clears out the board so that the game can be played again.
	 */
	public void resetBoard()
	{
		board = new Color[rows][cols];
	}
	
	/**
	 * This method allows for the object to be read with serialization
	 * @param aInputStream The input stream to use
	 */
	private void readObject(ObjectInputStream aInputStream) throws IOException, ClassNotFoundException
	{
     aInputStream.defaultReadObject();
  	}
	
	/**
	 * This method allows for the object to be written with serialization
	 * @param aOutputStream The output stream to use
	 */
	private void writeObject(ObjectOutputStream aOutputStream) throws IOException 
	{
      aOutputStream.defaultWriteObject();
   }
}