import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.Random;

/**
 * This class represents a Panel through which the user will interact with the
 * game. Holds references to the PlayingFields and an active Piece.
 */
public class PlayingPanel extends JPanel
{
	// Constants
	public static final int NUMBER_ROWS = 22;
	public static final int NUMBER_COLS = 10;
	public static final int FREE_ZONE_SIZE = 2;
	public static final int INIT_MOVE_TIME_MS = 800;
	public static final int MOVE_CHANGE_TIME_MS = 50;
	public static final int TILE_WIDTH = 25;
	public static final int TILE_HEIGHT = 25;
	public static final int ROW_VALUE = 100;
	public static final int SCORE_TO_LEVEL = 2000;

	// Class varaibles
	private Random	rand = new Random();

	// Instance variables
	private PlayingField field;
	private Piece activePiece;
	private Piece nextPiece;
	private Timer fallTimer;
	private Dimension tileDimension;
	private int score;
	private int level;
	private int moveTimeMs;
	private SLPanel scoreLevelPanel;
	private boolean paused;
	
	/**
	 * Constructor. Creates an new instance of PlayingPanel.
	 */
	public PlayingPanel()
	{	
		// Set to Border Layout
		setLayout(new BorderLayout());
		
		// Set background color
		setBackground(Color.BLACK);
	
		// Create new playing field
		field = new PlayingField(NUMBER_ROWS, NUMBER_COLS, FREE_ZONE_SIZE);
		
		// Set initial score and level
		score = 0;
		level = 1;
		
		// Create timer object
		moveTimeMs = INIT_MOVE_TIME_MS;
		fallTimer = new Timer(moveTimeMs, new TimerListener());
		fallTimer.start();
		
		// Set to unpaused
		paused = false;
		
		// Create tile dimension object
		tileDimension = new Dimension(TILE_WIDTH, TILE_HEIGHT);
		
		// Create a new active and next pieces
		activePiece = createNewPiece();
		nextPiece = createNewPiece();
		
		// Set SLPanel object
		scoreLevelPanel = new SLPanel(score, level, nextPiece, this);
		
		// Set to correct size
		setPreferredSize(new Dimension(NUMBER_COLS*TILE_WIDTH, NUMBER_ROWS*TILE_HEIGHT+100));
		doResize();
		
		// Set keyboard focus
		setFocusable(true);
		
		// Add keyboard listener
		addKeyListener(new KeyBoardListener());
		
		// Add property change listener
		addComponentListener(new SizeChangeListener()); 
	}
	
	/**
	 * Getter for PlayingField
	 * @return PlayingField object
	 */
	public PlayingField getField()
	{
		return field;
	}	
	
	/**
	 * Getter for active piece
	 * @return The active piece
	 */
	public Piece getActivePiece()
	{
		return activePiece;
	}
	
	/**
	 * Getter for next piece
	 * @return The next piece
	 */
	public Piece getNextPiece()
	{
		return nextPiece;
	}
	
	/**
	 * Getter for the score
	 * @return The current score
	 */
	public int getScore()
	{
		return score;
	}
	
	/**
	 * Getter for the level
	 * @return The current level
	 */
	public int getLevel()
	{
		return level;
	}
			
	/**
	 * Getter for the current timer
	 * @return The current timer delay
	 */
	public int getMoveTimeMs()
	{
		return moveTimeMs;
	}
	
	/**
	 * Getter for paused
	 * @return Whether or not it is paused
	 */
	public boolean getPaused()
	{	
		return paused;
	}
		
	/**
	 * Method for drawing all of the pieces and tiles.
	 * @param g The Graphics object
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		
		// Draw the field and the active Piece
		field.draw(g, tileDimension);
		activePiece.draw(g, tileDimension);
	}
	
	/**
	 * This method randomly creates a new Random Piece and returns it. 
	 * @return New Piece object
	 */
	private Piece createNewPiece()
	{
		Piece newPiece;
		// Choose piece and color randomly
		int pieceChoice = rand.nextInt(7);
		int colorChoice = rand.nextInt(7);
		
		// Create new color
		Color newColor;
		switch (colorChoice)
		{
			case 0: 
				newColor = Color.MAGENTA;
				break;
			case 1:
				newColor = Color.BLUE;
				break;
			case 2:
				newColor = Color.CYAN;
				break;
			case 3:
				newColor = Color.GREEN;
				break;
			case 4:
				newColor = Color.ORANGE;
				break;
			case 5:
				newColor = Color.RED;
				break;
			case 6:
				newColor = Color.YELLOW;
				break;
			default:
				newColor = Color.RED;
				break;
		}
		
		// Create new Piece for activePiece
		switch (pieceChoice)
		{
			case 0:
				newPiece = new LinePiece(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			case 1:
				newPiece = new LPiece1(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			case 2:
				newPiece = new LPiece2(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			case 3:
				newPiece = new SquarePiece(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			case 4:
				newPiece = new ZPiece1(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;	
			case 5:
				newPiece = new TPiece(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			case 6:
				newPiece = new ZPiece2(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
			default:
				newPiece = new ZPiece2(Math.round(NUMBER_COLS/2)-1, NUMBER_ROWS-1, newColor, field);
				break;
		}
		return newPiece;
	}
	
	/**
	 * This method is called when the game is lost. Asks the user if
	 * they want to play another game.
	 */
	private void resetGame()
	{
	 	// Let the user know that they lost
	 	JOptionPane.showMessageDialog(null, "I'm sorry, you lost. Press OK to play again", "Game Over", JOptionPane.PLAIN_MESSAGE);
		
		// Reset board and create a new piece
		field.resetBoard();
		activePiece = createNewPiece();
		nextPiece = createNewPiece();
		
		// Set score and level
		score = 0;
		level = 1;
		scoreLevelPanel.updateSL(score, level, nextPiece);
		
		// Reset timer
		moveTimeMs = INIT_MOVE_TIME_MS;
		fallTimer.setDelay(moveTimeMs);
	}
	
	/** 
	 * This class calculated the new size for everything and readds everything
	 * to the panel
	 */
	private void doResize()
	{
		// Removes components from panel
		removeAll();
		// Get the panel's dimension
		Dimension dim = getSize();
		// Cut out room for the bottom panel
		dim.height -= 100;
		// Change tile dimension
		tileDimension = new Dimension(dim.width/NUMBER_COLS, dim.height/NUMBER_ROWS);
		// Add space for tiles and readd bottom panel
		repaint();
		add(scoreLevelPanel, BorderLayout.SOUTH);
	}
	
	/**
	 * This method pauses the game
	 */
	public void pause()
	{
		if (!paused)
		{
			fallTimer.stop();
			paused = true; 
		}
	}
	
	/**
	 * This method unpaused the game
	 */
	public void unpause()
	{
		if(paused)
		{
			fallTimer.start();
			paused = false;
		}
	}
	
	/** 
	 * This method is invoked when a game is loaded from file. Sets all of the correct
	 * fields.
	 * @param loadedPanel The panel loaded from file
	 */
	public void load(PlayingPanel loadedPanel)
	{
		field = loadedPanel.getField();
		activePiece = loadedPanel.getActivePiece();
		nextPiece = loadedPanel.getNextPiece();
		score = loadedPanel.getScore();
		level = loadedPanel.getLevel();
		moveTimeMs = loadedPanel.getMoveTimeMs();
		fallTimer.setDelay(moveTimeMs);
	}
	
	/**
	 * Private class for listening to timer events
	 */ 
	private class TimerListener implements ActionListener
	{
		/**
		 * Method invoked after each amount of time determined by timer.
		 * Moves the active piece down one square, or settles the piece
		 * if that is not possible.
		 * @param e ActionEvent object
		 */
		public void actionPerformed(ActionEvent e)
		{
			// Move the piece down one row
			Point curLocation = activePiece.getLocation();
			curLocation.y--;
			boolean moveSuccessful = activePiece.move(curLocation);
	
			if (!moveSuccessful)
			{
				// If the move was not successful, settle the piece and create a new one
				activePiece.settle();
				int clearedNum = field.checkForClears();
				score += level*(clearedNum*clearedNum*ROW_VALUE);
				
				// Update the level and timer
				level = score / SCORE_TO_LEVEL + 1;
				moveTimeMs = INIT_MOVE_TIME_MS - (MOVE_CHANGE_TIME_MS * (level-1));
				if (moveTimeMs < 100)
					moveTimeMs = 100;
				fallTimer.setDelay(moveTimeMs); 
				
				if (field.isLossConditionMet())
				{
					resetGame();
				}
				else
				{
					activePiece = nextPiece;
					nextPiece = createNewPiece();
				}
			}
			
			// Repaint the panel
			repaint();
			
			// Update score level and next piece information
			scoreLevelPanel.updateSL(score, level, nextPiece);
		}
	}	
	
	/**
	 * Private class for reacting to user keyboard events (left right and space)
	 */
	private class KeyBoardListener implements KeyListener
	{
		/**
		 * Invoked when user presses a key. Moves piece left or right if possible, or
		 * rotates piece on space if possible.
		 * @param e KeyEvent object to get information about the specific event triggered
		 */
		public void keyPressed(KeyEvent e)
		{
			// Only works when the panel is not paused
			if (paused)
				return;
			// Get the key code
			int keyCode = e.getKeyCode();
			Point curLocation = activePiece.getLocation();
			switch (keyCode)
			{
				case KeyEvent.VK_LEFT:
					// Move the piece left one column
					curLocation.x--;
					activePiece.move(curLocation);
					break;
				case KeyEvent.VK_RIGHT:
					// Move the piece right one column
					curLocation.x++;
					activePiece.move(curLocation);
					break;
				case KeyEvent.VK_DOWN:
					// Move the piece down one row
					curLocation.y--;
					activePiece.move(curLocation);
					break;
				case KeyEvent.VK_SPACE:
					// Rotate the piece
					activePiece.rotate();
					break;
			}
			
			// Repaint the panel
			repaint();
		}
		
		/**
		 * Stubbed method for KeyListener interface
		 * @param e KeyEvent object
		 */
		public void keyReleased(KeyEvent e) {}
		
		/**
		 * Stubbed method for KeyListener interface
		 * @param e KeyEvent object
		 */
		public void keyTyped(KeyEvent e) {}
	}
	
	/**
	 * Private class for reacting to user resizing events
	 */
	private class SizeChangeListener implements ComponentListener
	{
		/**
		 * Method invoked when a panel is resized. Handles resizing.
		 */
		public void componentResized(ComponentEvent e)
		{
			doResize();
		}
		
		/** 
		 * Stubbed out methods
		 */
		public void componentHidden(ComponentEvent e) {}
		public void componentMoved(ComponentEvent e) {}
		public void componentShown(ComponentEvent e) {} 
	}
}